@@ -66,7 +66,7 @@ JackAudioInterface::JackAudioInterface(JackTrip* jacktrip,
                                        int NumNetRevChans,
                                        #endif // endwhere
                                        AudioInterface::audioBitResolutionT AudioBitResolution,
-                                       const char* ClientName) :
+                                       QString ClientName) :
     AudioInterface(jacktrip,
                    NumInChans, NumOutChans,
                    #ifdef WAIR // wair
@@ -81,6 +81,7 @@ JackAudioInterface::JackAudioInterface(JackTrip* jacktrip,
     mBitResolutionMode(AudioBitResolution),
     mClient(NULL),
     mClientName(ClientName),
+    mBroadcast(false),
     mJackTrip(jacktrip)
 {}
 
@@ -102,8 +103,23 @@ void JackAudioInterface::setup()
 //*******************************************************************************
 void JackAudioInterface::setupClient()
 {
-    const char* client_name = mClientName;
-    const char* server_name = NULL;
+    QByteArray clientName = mClientName.toUtf8();
+//    const char* server_name = NULL;
+#ifdef __MAC_OSX__
+    //Jack seems to have an issue with client names over 27 bytes in OS X
+    int maxSize = 27;
+#else
+    int maxSize = jack_client_name_size();
+#endif
+    if (clientName.length() > maxSize) {
+        int length = maxSize;
+        //Make sure we don't cut mid multi-byte character.
+        while ((length > 0) && ((clientName.at(length) & 0xc0) == 0x80)) {
+            length--;
+        }
+        clientName.truncate(length);
+    }
+
     // was  jack_options_t options = JackNoStartServer;
     // and then jack_options_t options = JackLoadName;
     jack_options_t options = JackNullOption; // from jackSimpleClient example
@@ -114,10 +130,15 @@ void JackAudioInterface::setupClient()
     /// verbose message, check how to desable them.
     {
         QMutexLocker locker(&sJackMutex);
+//#ifndef WAIR // WAIR
+//        mClient = jack_client_open (client_name, options, &status, server_name);
+//#else
+//        mClient = jack_client_open (client_name, JackUseExactName, &status, server_name);
+//#endif // endwhere
 #ifndef WAIR // WAIR
-        mClient = jack_client_open (client_name, options, &status, server_name);
+        mClient = jack_client_open (clientName.constData(), options, &status);
 #else
-        mClient = jack_client_open (client_name, JackUseExactName, &status, server_name);
+        mClient = jack_client_open (clientName.constData(), JackUseExactName, &status);
 #endif // endwhere
     }
 
@@ -136,8 +157,7 @@ void JackAudioInterface::setupClient()
         fprintf (stderr, "JACK server started\n");
     }
     if (status & JackNameNotUnique) {
-        client_name = jack_get_client_name(mClient);
-        fprintf (stderr, "unique name `%s' assigned\n", client_name);
+        fprintf (stderr, "unique name `%s' assigned\n", jack_get_client_name(mClient));
     }
 
     // Set function to call if Jack shuts down
@@ -152,6 +172,7 @@ void JackAudioInterface::setupClient()
     // Initialize Buffer array to read and write audio
     mInBuffer.resize(mNumInChans);
     mOutBuffer.resize(mNumOutChans);
+    mBroadcastBuffer.resize(mNumOutChans);
 }
 
 
@@ -179,6 +200,18 @@ void JackAudioInterface::createChannels()
                                            JACK_DEFAULT_AUDIO_TYPE,
                                            JackPortIsOutput, 0);
     }
+    //Create Broadcast Ports
+    if (mBroadcast) {
+        mBroadcastPorts.resize(mNumOutChans);
+        for (int i = 0; i < mNumInChans; i++)
+        {
+            QString outName;
+            QTextStream (&outName) << "broadcast_" << i+1;
+            mBroadcastPorts[i] = jack_port_register (mClient, outName.toLatin1(),
+                                               JACK_DEFAULT_AUDIO_TYPE,
+                                               JackPortIsOutput, 0);
+        }
+    }
 }
 
 
@@ -254,7 +287,9 @@ int JackAudioInterface::stopProcess() const
 void JackAudioInterface::jackShutdown (void*)
 {
     //std::cout << "The Jack Server was shut down!" << std::endl;
-    throw std::runtime_error("The Jack Server was shut down!");
+    JackTrip::sJackStopped = true;
+    std::cout << "The Jack Server was shut down!" << std::endl;
+    //throw std::runtime_error("The Jack Server was shut down!");
     //std::cout << "Exiting program..." << std::endl;
     //std::exit(1);
 }
@@ -264,10 +299,15 @@ void JackAudioInterface::jackShutdown (void*)
 //*******************************************************************************
 int JackAudioInterface::processCallback(jack_nframes_t nframes)
 {
+  if(mProcessingAudio) {
+    std::cerr << "*** JackAudioInterface.cpp: DROPPED A BUFFER because AudioInterface::callback() not finished\n";
+    return 1;
+  }
+
     // Get input and output buffers from JACK
     //-------------------------------------------------------------------
     for (int i = 0; i < mNumInChans; i++) {
-        // Input Ports are READ ONLY
+        // Input Ports are READ ONLY and change as needed (no locks) - make a copy for debugging
         mInBuffer[i] = (sample_t*) jack_port_get_buffer(mInPorts[i], nframes);
     }
     for (int i = 0; i < mNumOutChans; i++) {
@@ -283,6 +323,14 @@ int JackAudioInterface::processCallback(jack_nframes_t nframes)
     //-------------------------------------------------------------------
 
     AudioInterface::callback(mInBuffer, mOutBuffer, nframes);
+
+    if (mBroadcast) {
+        for (int i = 0; i < mNumOutChans; i++) {
+            // Broadcast Ports are WRITABLE
+            mBroadcastBuffer[i] = (sample_t*) jack_port_get_buffer(mBroadcastPorts[i], nframes);
+        }
+        AudioInterface::broadcastCallback(mBroadcastBuffer, nframes);
+    }
     return 0;
 }
 
@@ -358,7 +406,7 @@ void JackAudioInterface::connectDefaultPorts()
 
 
 
-// OLD CODE
+// OLD CODE (some moved to parent class AudioInterface.cpp)
 // ==============================================================================
 
 //*******************************************************************************
